MODULE DhryxSPC;
(*                                                                         .
   Version 1.3.RS * 03.02.94 ~ 26.03.14: Anpassung an TDI, HM2, LPR & METH
    Die Unterbereichstypen muten fr TDI leider durch Integer ersetzt
     werden, sonst Meldungen: " Compiler Error in Assert "!     .

 * Compilercode-Testprogramm.
 * Version 1.2, 15.05.90
 *   Portiert durch Dirk Steins von C-Source (Version 1.1) nach Modula-2,
 * berarbeitung & Dokumentation von Thomas Tempelmann.
 *
 *   Register-Variable werden nicht im einzelnen spezifiziert, da dies
 * bei Modula-2 nicht vorgesehen ist.
 *
 *   Dafr wird in 'Proc1' eine WITH-Anweisung verwendet, fr die es in
 * der C-Version nur ein Macro gibt (C bietet kein WITH-Konstrukt). Da
 * aber alles darauf hinweist, da im Original dieses Programms, das in
 * ADA geschrieben wurde, WITH verwendet wurde, wird es auch hier in der
 * Modula-Version getan, schon allein, weil es sinnvoll fr den Test ist.
 *
 *   Fr den Test sollten alle globalen Optimierungen aktiviert sein
 * und Bereichs-, berlauf-, Stack- und weitere Prfungen deaktiviert sein.
 *
 * ACHTUNG:
 *   Um korrekte Ergebnisse zu erhalten, sollten alle residenten Programme,
 * also Programme im AUTO-Ordner und Accessories, beim Test entfernt werden,
 * weil viele dieser Programme sich in System-Vektoren installieren (zB VBL)
 * und damit den Rechner ein wenig verlangsamen.
 *   Auerdem ist zu beachten, da auf manchen Systemen (zB. Falcon) das
 * Video-System den Bus belastet: Dann also immer mit der gleichen Auflsung
 * (ie. ST-HI) testen! (Oder sogar Video abschalten?)
 *
 * Ergebnisse:
 *
 * Compiler       Hardware   MHz  Dhrystones/s
 * =========================================================
 * Megamax 3.8     520ST+     8       522
 * FTL 1.18        520ST+     8       656   (s. Anm.1)
 * SPC 2.0         520ST+     8     493/683/855 (s. Anm.2,3)
 * Hnisch 3.105   520ST+     8       710
 * TDI 3.01        520ST+     8       717
 * DRI-AlcyonC4.14 MegaST4    8       719 ? (PD338)
 * M2AMIGA 3.3     Amiga2000  7.16    847 (2MB FastRAM)
 * LPR 1.4.RS      MegaST4    8       953 (s. Anm.3)
 * ETH 2.RS        MegaST4    8       955 (s. Anm.3)
 * FST 2.0         IBM286AT  12      1060
 * Megamax 4.0     520ST+     8      1061
 * TDI 4.RS        MegaST4    8      1062
 * HM2 5.10,RS     MegaST4    8    1090/1183 (s. Anm.3)
 * ASM?            ..ST..     8      1105 (Referenz aus FastTech.PRG)
 * TDI 4.RS        Falcon     8/8    1250 (Cache!)
 * Rowley 1.34     520ST+     8      1500
 * TDI 4.RS        MegaSTe   16/8    1814
 * TDI 4.RS        MegaST4   20/8    ????
 * ETH 2.RS        Falcon    32/16   2525 (ddd)
 * TDI 4.RS        Falcon    16/16   2551
 * TDI 4.RS        Falcon    32/16   2773 (SpeedRes)
 * TDI 4.RS        Falcon    18/18   2840 (SpeedRes)
 * TDI 4.RS        Falcon    32/16   2911 (ddd)
 * TDI 4.RS        Falcon    32/16   3033 (ddd, ohne Video)
 * TDI 4.RS        Falcon    36/18   3139 (SpeedRes)
 * TDI 4.RS        Falcon    20/20   3174 (SpeedRes)
 * TDI 4.RS        Falcon    36/20?  3478 (SpeedRes)
 * TDI 4.RS        Falcon    40/20   3521 (SpeedRes)
 * LogiTech 3.4DOS 80386     24/0ws  4755
 * Rowley 1.34     Sun3/60   20      5500 (68020er)
 * TDI 4.RS        TT30      32/16  ~6000 (gemittelt; 32er-Bus! FPU/50)
 * Rowley 1.33     T800      20      8500
 * Rowley 1.33     R3000     16     12500
 *
 * Anmerkungen zu den Ergebnissen:
 *
 *  1) Der FTL-Compiler verfgt ber keine "Compare"-Funktion im Strings-
 *     Modul, daher wurde die eigens dafr erstellte Routine hier im
 *     Testprogramm (su.) benutzt.
 *  2) Einige Modula-Systeme bieten eine sehr langsame "Compare"-Funktion.
 *     Da man sie selbst womglich gegen eine schnellere, wie die unten
 *     stehende, "Compare"-Funktion austauschen wrde, wurden bei diesen
 *     Systemen zwei Ergebnisse ermittelt: mit der Strings-Funktion und
 *     mit der hiesigen Funktion. Zu beachten ist aber, da die hiesige
 *     Funktion mit VAR-Parametern arbeitet, soda sie nicht allgemein
 *     einsetzbar ist. Dies sollte aber - wegen des Gedankens der eigenen
 *     Optimierung, so beibehalten bleiben!
 *  3) Mit einem per CODE optim. Compare a la SX.CMP geht die Post ab!
 *     Da ein schnelles Compare so leicht nachzursten geht, fragt man sich,
 *     was es berhaupt in diesem Benchmark zu suchen hat; jedenfalls sollte
 *     das kein ernsthaftes Kriterium sein! RS
 *  4) Dhry als Hardware-Vergleich:
 *     Anscheinend hat der Bus-Takt mehr Einflu auf Dhry als der CPU-Takt!
 *     Und an FPU, Videl, DSP &c. geht der Dhry vorbei. RS
 *)

(***** Compiler-/Library-abhng. Importe, Direktiven, Def. & Funktionen ****)

FROM   SYSTEM  IMPORT (* CODE,*) INLINE, VAL, REG, (**) ADR;
(* FROM String IMPORT (* CompareResults,*) Compare; *)
IMPORT InOut; (* Aux, VT,VN,*) (* SX; *)

   (*$Q+ TDI:    local quick bras *)
   (* A+ TDI:   global quick bras *)
   (*$T- TDI:     no range checks *)
   (*$R- Megamax: no range checks *)
   (*$S- TDI, " : no stack checks *)
   (* Z+ Megamax: optimize for use of registers in function returns *)
   (* I+ FTL: HIGH() returns CARDINAL (16 Bit) *)
   (*$V- METH:   no overflow chex.*)

CONST Version =  "1.3RS/SPC";
      LOOPS = 10000;  (* fr ca. 10~20 sec bei 8MHz *)

TYPE  IG = (* SHORTINT  *) INTEGER ; (* 16 bit type *)
      CD = (* SHORTCARD *) CARDINAL; (* 16 bit type *)

      IG1to30 = IG (**) [1..30] (**);
      IG1to50 = IG (**) [1..50] (**);

CONST PM = 4 +4 +4;         (* ParameterMark (frameoffset): RetAdr +A4 +SP.*)
CONST Greater = 1; Equal = 0; Less = -1;

TYPE  Rel = IG; (* SX.REL; *) (* CompareResults; *)
      CmpProc =
        PROCEDURE (VAR ARRAY OF CHAR, VAR (**) ARRAY OF CHAR (*, CD *)): Rel;
VAR   Compara: CmpProc;

PROCEDURE CMP (VAR L, R: ARRAY OF CHAR): INTEGER;  (* P-*)  (* Nur FIN =0c.*)
 BEGIN INLINE ( 206FH,PM+6, 226FH,PM, 1018H, 6700H+6, 9019H, 6800H-8 );
       INLINE ( 6000H+2, 9019H, 4880H (*, 3F40H,PM+12, 4E75H *) );
   RETURN VAL (INTEGER, REG(0)); END CMP; (* P+*)

TYPE  LC = LONGCARD;
      sProc = PROCEDURE ((* VAR *) ARRAY OF CHAR);
      cProc = PROCEDURE ((* LC *) CD, CD);
VAR   wS: sProc; wC: cProc; Ln: PROC;

PROCEDURE Init;
 BEGIN Compara := (* Compare *) (* SX.*) CMP;
   Ln := InOut.WriteLn (* VT.Ln *);
   wS := InOut.WriteString (* sProc (VT.wT) *);
   wC := InOut.WriteCard   (* cProc (VN.wDecW) *);
   wS (33C); wS ("H Running... ");
   Ln; END Init;

PROCEDURE Pause; VAR ch: CHAR;
 BEGIN InOut.Read (ch); (* VN.Pause; *) END Pause;

CONST HZ200 = 4BAH; HZ = 200D;  (* time() RETURNs 1/200 seconds (Atari ST).*)
VAR   hz200: LC;
      Hz200: RECORD CASE: CHAR OF 0C: p: POINTER TO LC ELSE c: LC END END;

(*$S- TDI*) (*$E+ HM2*)
PROCEDURE readTimer; BEGIN hz200 := Hz200.p^ END readTimer; (*$S=,$E=*)
PROCEDURE SuperExec (Code: PROC);
 BEGIN INLINE (2F2EH,PM, 3F3CH,38, 4E40H+14(*XBIOSTRAP*)) END SuperExec;
PROCEDURE time(): LC;  (* Diese Funktion liest den 200-Hz-Timer des ST aus.*)
 BEGIN Hz200.c := HZ200; SuperExec (readTimer); RETURN hz200 END time;

PROCEDURE wResult (benchtime: LC);
   VAR l: LC;
 BEGIN l := LOOPS; Ln;
   wS (" Modula-2 Dhrystone ("); wS (Version);
   wS (") time for "); wC (LOOPS, 1); wS (" passes is: ");
(* wC ( CD (Aux.DivI (benchtime, HZ)), 1); *)
   wC (benchtime DIV HZ, 1);
   wS (" sec."); Ln;
   wS (" This system benchmarks at ");
(* wC ( CD (Aux.DivI (Aux.MulI (LC (LOOPS), HZ), CD (benchtime))), 1); *)
   wC ((l * HZ) DIV benchtime, 1);
   wS (" dhrystones/sec. "); END wResult;

(* Fr Systeme, die keine Compare-Funktion haben (zB. FTL),
                  oder eine DEUTLICH zu langsame (zB. SPC):

PROCEDURE Compare (VAR left, right: ARRAY OF CHAR): Rel;
       (* Die VAR-Parameter sind fr eine individuell auf diese
        * Testanwendung erstellte Funktion legitim. Dafr
        * bekommt das Modula-System aber auch gleich Minuspunkte,
        * weil es diese Funktion nicht bereitstellt! *)
 VAR   idx, high: IG; ch: CHAR;
 BEGIN IF  (HIGH (left) > HIGH (right)) THEN
            high := HIGH (right)
       ELSE high := HIGH (left) END;
       idx := 0;
   REPEAT ch := left[idx];
     IF  (ch # right[idx]) THEN
      IF (ch > right[idx]) THEN RETURN Greater ELSE RETURN Less END END;
     IF  (ch = 0c) THEN RETURN Equal END;
     INC (idx)
   UNTIL (idx > high);
   IF (HIGH (left) = HIGH (right)) THEN RETURN Equal END;
   IF (HIGH (left) > HIGH (right)) THEN
        IF (left [idx] = 0c) THEN RETURN Equal ELSE RETURN Greater END;
   ELSE IF (right[idx] = 0c) THEN RETURN Equal ELSE RETURN Less END END;
   END Compare;
*)

(***** Beginn des Compiler-/Library-unabhngigen Programms *****)

TYPE OneToThirty = IG1to30;
TYPE OneToFifty  = IG1to50;
TYPE Enumeration = (Ident1, Ident2, Ident3, Ident4, Ident5);
TYPE CapLetter   = CHAR;
TYPE String30    = ARRAY [0..30] OF CHAR;
TYPE Array1Dim   = ARRAY [0..50] OF IG;
TYPE Array2Dim   = ARRAY [0..50],[0..50] OF IG;

TYPE RecordPtr   = POINTER TO RecordType;
     RecordType  = RECORD
       PtrComp   : RecordPtr;
       Discr,
       EnumComp  : Enumeration;
       IntComp   : OneToFifty;
       StringComp: String30;   END;

(* Package 1 *)
VAR
    IntGlob   : IG;
    BoolGlob  : BOOLEAN;
    Char1Glob,
    Char2Glob : CHAR;
    Array1Glob: Array1Dim;
    Array2Glob: Array2Dim;
    PtrGlb,
    PtrNext   : RecordPtr;
    GlbRec,
    NextRec   : RecordType;

PROCEDURE Func1 (CharPar1, CharPar2: CapLetter): Enumeration;
 VAR   CharLoc1,
       CharLoc2 : CapLetter;
 BEGIN CharLoc1 := CharPar1;
       CharLoc2 := CharLoc1;
   IF (CharLoc2 # CharPar2) THEN RETURN Ident1 ELSE RETURN Ident2 END;
   END Func1;

PROCEDURE Func2 ( VAR StrParI1, StrParI2: String30): BOOLEAN;
 VAR   IntLoc:  OneToThirty;
       CharLoc: CapLetter;
 BEGIN IntLoc := 1;
   WHILE (IntLoc <= 1) DO
     IF (Func1 (StrParI1[IntLoc], StrParI2[IntLoc+1]) = Ident1) THEN
       CharLoc := 'A'; INC (IntLoc); END END;
   IF (CharLoc >= 'W') & (CharLoc < 'Z') THEN IntLoc := 7 END;
   IF (CharLoc  # 'X') THEN
     IF (Compara (StrParI1, StrParI2 (*, 0*)) >= Equal) THEN INC (IntLoc, 7);
     ELSE RETURN FALSE END; END;
   RETURN TRUE; END Func2;

PROCEDURE Func3 (EnumParIn: Enumeration): BOOLEAN;
 VAR   EnumLoc:  Enumeration;
 BEGIN EnumLoc:= EnumParIn;
   IF (EnumLoc = Ident3) THEN RETURN TRUE END;
   RETURN FALSE; END Func3;

PROCEDURE Proc7 (IntParI1, IntParI2: OneToFifty; VAR IntParOut: OneToFifty);
 VAR   IntLoc: OneToFifty;
 BEGIN IntLoc := IntParI1 +2; IntParOut := IntParI2 + IntLoc END Proc7;

PROCEDURE Proc3 (VAR PtrParOut: RecordPtr);
 BEGIN IF (PtrGlb # NIL) THEN PtrParOut := PtrGlb^.PtrComp;
       ELSE   IntGlob := 100 END;
   Proc7 (10, IntGlob, PtrGlb^.IntComp); END Proc3;

PROCEDURE Proc6 (EnumParIn: Enumeration; VAR EnumParOut: Enumeration);
 BEGIN EnumParOut := EnumParIn;
   IF (~ Func3 (EnumParIn)) THEN EnumParOut := Ident4 END;
   CASE EnumParIn OF
   | Ident1: EnumParOut := Ident1;
   | Ident2: IF (IntGlob > 100) THEN EnumParOut := Ident1
             ELSE EnumParOut := Ident4 END;
   | Ident3: EnumParOut := Ident2;
   | Ident4:
   | Ident5: EnumParOut := Ident3;
   ELSE END; END Proc6;

PROCEDURE Proc1  (PtrParIn: RecordPtr);
 BEGIN  PtrParIn^.PtrComp^ := PtrGlb^;
        PtrParIn^.IntComp := 5;
   WITH PtrParIn^.PtrComp^ DO
     IntComp := PtrParIn^.IntComp;
     PtrComp := PtrParIn^.PtrComp;
     Proc3 (PtrComp);
     IF (Discr = Ident1) THEN IntComp := 6;
        Proc6 (PtrParIn^.EnumComp, EnumComp);
        PtrComp := PtrGlb^.PtrComp;
        Proc7 (IntComp, 10, IntComp);
     ELSE PtrParIn := PtrParIn^.PtrComp END;
     END; END Proc1;

PROCEDURE Proc2 (VAR IntParIO: OneToFifty);
 VAR   IntLoc : OneToFifty;
       EnumLoc: Enumeration;
 BEGIN IntLoc := IntParIO +10;
   LOOP IF (Char1Glob = 'A') THEN DEC (IntLoc);
            IntParIO := IntLoc - IntGlob;
            EnumLoc  := Ident1; END;
        IF (EnumLoc = Ident1) THEN EXIT END;
        END; END Proc2;

PROCEDURE Proc4 ();
 VAR   BoolLoc : BOOLEAN;
 BEGIN BoolLoc := Char1Glob = 'A';
       BoolLoc := ~ BoolGlob;
       Char2Glob := 'B'; END Proc4;

PROCEDURE Proc5 ();
 BEGIN Char1Glob := 'A'; BoolGlob := FALSE END Proc5;

PROCEDURE Proc8 (VAR Array1Par: Array1Dim; VAR Array2Par: Array2Dim;
                 IntParI1, IntParI2: OneToFifty);
 VAR   IntLoc, IntIndex: OneToFifty;
 BEGIN IntLoc := IntParI1 + 5;
   Array1Par[IntLoc] := IntParI2;
   Array1Par[IntLoc+1] := Array1Par[IntLoc];
   Array1Par[IntLoc+30] := IntLoc;
   FOR IntIndex := IntLoc TO IntLoc+1 DO
       Array2Par[IntLoc][IntIndex] := IntLoc; END;
   INC(Array2Par[IntLoc][IntLoc-1]);
   Array2Par[IntLoc+20][IntLoc] := Array1Par[IntLoc];
   IntGlob := 5; END Proc8;

PROCEDURE Proc0 ();
 VAR   IntLoc1,
       IntLoc2,
       IntLoc3   : OneToFifty;
       CharLoc,
       CharIndex : CHAR;
       EnumLoc   : Enumeration;
       String1Loc,
       String2Loc: String30;
       starttime,
       nulltime  : LC;
       i         : [0..LOOPS];
 BEGIN starttime := time();
       FOR i := 0 TO LOOPS-1 DO END;
       nulltime  := time() -starttime;        (* Computes overhead of loop.*)
(* NEW (PtrNext); *)
   PtrNext := ADR (NextRec);
   PtrGlb  := ADR (GlbRec);
(* NEW (PtrGlb); *)
   PtrGlb^.PtrComp := PtrNext;
   PtrGlb^.Discr   := Ident1;
   PtrGlb^.EnumComp:= Ident3;
   PtrGlb^.IntComp := 40;
   PtrGlb^.StringComp
               := "DHRYSTONE PROGRAM, SOME STRING";
   String1Loc  := "DHRYSTONE PROGRAM, 1'ST STRING";  (*GOOF*)
   Array2Glob[8][7] := 10;
(***** Start Timer: *****)
   starttime := time();
   FOR i := 0 TO LOOPS-1 DO
     Proc5;
     Proc4;
     IntLoc1 := 2;
     IntLoc2 := 3;
     String2Loc:= "DHRYSTONE PROGRAM, 2'ND STRING";
     EnumLoc := Ident2;
     BoolGlob := ~ Func2 (String1Loc, String2Loc);
     WHILE (IntLoc1  < IntLoc2) DO
       IntLoc3 :=  5 * IntLoc1 -IntLoc2;
       Proc7 (IntLoc1, IntLoc2, IntLoc3);
       INC (IntLoc1); END;
     Proc8 (Array1Glob, Array2Glob, IntLoc1, IntLoc3);
     Proc1 (PtrGlb);
     FOR CharIndex := 'A' TO Char2Glob DO
       IF (EnumLoc = Func1 (CharIndex, 'C')) THEN
          Proc6 (Ident1, EnumLoc); END; END;
     IntLoc3 := IntLoc2  *  IntLoc1;
     IntLoc2 := IntLoc3 DIV IntLoc1;
     IntLoc2 := 7 * (IntLoc3 -IntLoc2) -IntLoc1;
     Proc2 (IntLoc1); END;
(***** Stop Timer: *****)
   wResult (time() -starttime -nulltime);
   END Proc0;

BEGIN Init; Proc0;
  Pause; END DhryxSPC .
